/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CmdLine is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* userFunc to kick isakmp in the pants/
/* Use the isakmp.cap with this userFunc - it's a MainMode-quickMode capture*/
/* When fuzzing iskampd with -b and this user func, you'll probably want to use 43 as AttackChar so the currentPkt is randomized */
/* When using -l use any byte and try 43 also */
/* Guess it wouldn't hurt to try -f, but it seems unlikely there would be a format bug in a binary/encrypted protocol.*/

///Shared globals defined in isa.h.  Included in GPF.c
///ret=0 means ok, ret=1 means start the sequence again (goto first pkt); we only do that if not attack and only after 
///the second pkt

int isakmp(int readRet, unsigned int index, unsigned int location, unsigned int currentPosition, char * currentPkt, unsigned int currentPktLen)
{
	int ret=0, x;
/*Phase 1*/
/*All pkts we send out and early start up stuff here*/
	///generate a pesudorandom icookie and fill all outgoing pkts with it
	if (icookieConstruct == 0)
	{
		icookie = (rand() & 0xffffffff);
		icookieConstruct = 1;
		///determine if we'll properly use cookies (89%)
		if ( (rand()&0xff)>30 )
			goodCookie=1;
	}
	///put in icookie for all non-fuzzed pkts
	if( location==0)
		*((u_int64_t *)Data[index])=icookie;
	///For the sake of the -l option I'll clear the bytes in the rcookie position expect when currentPos is on them
	if ( ((index==0) && (location == 3) && !(currentPosition>=8 && currentPosition<=15)) )
		*((u_int64_t *)currentPkt+1)=0x0000000000000000;
	///if we're about to close the socket, reset icookieConstruct to 0 so we'll get a new icookie
	if (location == 4)
	{
		icookieConstruct=0;
		goodCookie=0;
	}

/*1st pkt(send) - don't forget numbering begins at 0 in most all of GPF. (i.e. index==0 == pkt1)*/
	///Set icookie in the case this pkt is the one being fuzzed (80%)
	if ( (index==0) && (location == 3) && (goodCookie) )
	{
		*((u_int64_t *)currentPkt)=icookie;
	}
	///fix up hdr (80% of the time) when hdr isn't being fuzzed (next payload, version, exchange type, flags, len)
	if(index==0 && location==3 && ((rand()&0xff) > 50) )
	{
		if ( !(currentPosition==16) )
			*(currentPkt+16)=0x01;
		if ( !(currentPosition==17) )
			*(currentPkt+17)=0x10;
		if ( !(currentPosition==18) )
			*(currentPkt+18)=0x02;
		if ( !(currentPosition==19) )
			*(currentPkt+19)=0x00;
		if ( !(currentPosition>=24 && currentPosition<=27) )
			*((u_int32_t *)currentPkt+6)=htonl(currentPktLen);
	}
	///fix up the lengths when we fuzz vendor_ids (80%)
	if ( (index==0) && (location==3) && ((rand()&0xff) >50) )
	{
		if(currentPosition>=83 && currentPosition<=96)
			*((u_int16_t *)currentPkt+41)=htons(currentPktLen-140);  ///orglen=160; vendoridlen=20; thus the 140
		else if (currentPosition>=103 && currentPosition<=116)
			*((u_int16_t *)currentPkt+51)=htons(currentPktLen-140);
		else if (currentPosition>=123 && currentPosition<=136)
			*((u_int16_t *)currentPkt+61)=htons(currentPktLen-140);
		else if (currentPosition>=143 && currentPosition<=156)
			*((u_int16_t *)currentPkt+71)=htons(currentPktLen-140);
	}

/*2nd pkt (recv)*/
	///This is how we'll SYNC for UDP
	if( (index==1) && (location==1) )
	{
        	if( (readRet!=-2) && (*((u_int64_t *)currentPkt)!=0) && (*(currentPkt+16)==0x01) )
		{
			badpkts--;
                	ret=0;
		}
        	else
		{
			icookieConstruct=0;
			badpkts++;
			if( badpkts>10)
			{	
				printf("Too many bad pkt from other end...lets sleep for a bit and hope that helps\n");
				sleep(2);
				badpkts=0;
			}
                	return 1;
		}
	}
	///snag the rcookie and store it, so we can properly use it later
	if ( (index==1) && (location == 1) )
	{
		rcookie=*((u_int64_t *)readData[index]+1);
		rcookieSet=1;
	}
	///if we're about to close the socket clear rcookie, so we'll snag the new one, next session
	if (location == 4)
		rcookieSet=0;
	///Sets rcookie in all nonfuzz cases
	if ( (index > 1) && (rcookieSet==1) && (location==0) )
		*((u_int64_t *)Data[index]+1)=rcookie;

/*3rd pkt (send)*/
	///Set rcookie & icookie in the case this pkt is the one being fuzzed (% based on previous set)
	if(goodCookie)
	{	
		if ( (rcookieSet==1) && (index==2) && (location == 3) )
			*((u_int64_t *)currentPkt+1)=rcookie;
		if ( (index==2) && (location == 3) )
			*((u_int64_t *)currentPkt)=icookie;
	}
	///fix up hdr (80% of the time) when hdr isn't being fuzzed (next payload, version, exchange type, flags, len)
	if(index==2 && location==3 && ((rand()&0xff) > 50) )
	{
		if ( !(currentPosition==16) )
			*(currentPkt+16)=0x04;
		if ( !(currentPosition==17) )
			*(currentPkt+17)=0x10;
		if ( !(currentPosition==18) )
			*(currentPkt+18)=0x02;
		if ( !(currentPosition==19) )
			*(currentPkt+19)=0x00;
		if ( !(currentPosition>=24 && currentPosition<=27) )
			*((u_int32_t *)currentPkt+6)=htonl(currentPktLen);
	}
	//Might as well fix up the key and nonce hdrs once and a while for fun (57%)
	if ( (index==2) && (location == 3) && ((rand()&0xff) > 110) )
        {	
		if (currentPosition==28)
			*(currentPkt+28)=0x0a;
		if (currentPosition==160)
			*(currentPkt+160)=0x00;
	}
	///we also want to fix up the length if we're fuzzing in key or nonce currentPkt positions (80%)
	if ( (index==2) && (location == 3) && ((rand()&0xff) > 50) )
        {
		if(currentPosition>=32 && currentPosition<=159)
                        *((u_int16_t *)currentPkt+15)=htons(currentPktLen-48);  //180-132=48
                else if (currentPosition>=164 && currentPosition<=179)
                        *((u_int16_t *)currentPkt+81)=htons(currentPktLen-160);  //180-20=160
        }

	///We want to properly use the key & nonce currentPkt, but I need to learn crypto stuff!!!

/*4th pkt (recv)*/
	

/*5th pkt (send)*/
	///Set rcookie & icookie in the case this pkt is the one being fuzzed (% based on previous set)
	if(goodCookie)
	{	
		if ( (rcookieSet==1) && (index==4) && (location == 3) )
			*((u_int64_t *)currentPkt+1)=rcookie;
		if ( (index==4) && (location == 3) )
			*((u_int64_t *)currentPkt)=icookie;
	}
	///fix up hdr (80% of the time) when hdr isn't being fuzzed (next payload, version, exchange type, flags, len)
	if(index==4 && location==3 && ((rand()&0xff) > 50) )
	{
		if ( !(currentPosition==16) )
			*(currentPkt+16)=0x05;
		if ( !(currentPosition==17) )
			*(currentPkt+17)=0x10;
		if ( !(currentPosition==18) )
			*(currentPkt+18)=0x02;
		if ( !(currentPosition==19) )
			*(currentPkt+19)=0x01;
		if ( !(currentPosition>=24 && currentPosition<=27) )
			*((u_int32_t *)currentPkt+6)=htonl(currentPktLen);
	}


/*6th pkt (recv)*/


/*Moving on to Phase 2*/
/*7th pkt*/

		
	
	return ret;
}
