/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CmdLine is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*This is the same as the other userFunc except it assumes that isakmpd will connect to us.*/
/*BE SURE TO USE THE CORRECT USERFUNC.  isakmp for client mode, isakmp_serv for server mode.*/

///Shared globles defined in isa.h.  included in GPF.c
///ret=0 means ok; ret=-1 means reread until we get what we're looking for

int isakmp_serv(int readRet, unsigned int index, unsigned int location, unsigned int currentPosition, char * currentPkt, unsigned int currentPktLen)
{
	int ret;
/*Phase 1*/
/*All pkts we send out and early start up stuff here*/
	///generate a pesudorandom rcookie and fill all outgoing pkts with it
	if(rcookieConstruct == 0)
	{
		rcookie = (rand() & 0xffffffff);
		rcookieConstruct = 1;
	}
	///put in rcookie for all non-fuzzed pkts
	if( location==0)
		*((u_int64_t *)Data[index]+1)=rcookie;
	///Determin if we'll use good cookies or not (89%)
	if ( (rand()&0xff)>30 )
		goodCookie=1;
	///if we're about to close the socket, reset rcookieConstruct to 0 so we'll get a new icookie
	if(location == 4)
	{
		rcookieConstruct=0;
		goodCookie=0;
	}

/*1st pkt(recv) - don't forget numbering begins at 0 in most all of GPF. (i.e. index==0 == pkt1)*/
	///This is how we'll SYNC for UDP
	if( (index==0) && (location==1) )
	{
		if( (readRet!=-2) && (*((u_int64_t *)currentPkt+1)==0) && (*(currentPkt+16)==0x01) ) //though we use p
			ret=0;
		else
		{
			rcookieConstruct=0;
			return -1;
		}
	}
	///snag the rcookie and store it, so we can properly use it later
	if( (index==0) && (location == 1) )
	{
		icookie=*((u_int64_t *)readData[index]);
		icookieSet=1;
	}
	///if we're about to close the socket clear rcookie, so we'll snag the new one, next session
	if(location == 4)
		icookieSet=0;
	///Sets rcookie in all nonfuzz cases
	if( (index > 0) && (icookieSet==1) && (location==0) )
		*((u_int64_t *)Data[index])=icookie;

/*2nd pkt (send)*/
	///Set rcookie in the case this pkt is the one being fuzzed (80%)
	if( (currentPkt!=NULL) && (index==0) && (location == 3) && (goodCookie) )
	{
		*((u_int64_t *)currentPkt+1)=rcookie;
	}
	///fix up hdr (80% of the time) when hdr isn't being fuzzed (next payload, version, exchange type, flags, len)
	if( (currentPkt!=NULL) && (index==0) && (location==3) && ((rand()&0xff) > 50) )
	{
		if( !(currentPosition==16) )
			*(currentPkt+16)=0x01;
		if( !(currentPosition==17) )
			*(currentPkt+17)=0x10;
		if( !(currentPosition==18) )
			*(currentPkt+18)=0x02;
		if( !(currentPosition==19) )
			*(currentPkt+19)=0x00;
		if( !(currentPosition>=24 && currentPosition<=27) )
			*((u_int32_t *)currentPkt+6)=htonl(currentPktLen);
	}
	///fix up the length when we fuzz vendor_id (80%)
	if( (currentPkt!=NULL) && (index==0) && (location==3) && ((rand()&0xff) >50) )
	{
		if(currentPosition>=83 && currentPosition<=96)
			*((u_int16_t *)currentPkt+41)=htons(currentPktLen-140);  ///orglen=160; vendoridlen=20; thus the 140
	}

/*3rd pkt (recv)*/

/*4th pkt (send)*/
	///Set rcookie & icookie in the case this pkt is the one being fuzzed (% based on previous set)
	if(goodCookie)
	{	
		if( (currentPkt!=NULL) && (icookieSet==1) && (index==3) && (location == 3) )
			*((u_int64_t *)currentPkt)=icookie;
		if( (index==3) && (location == 3) )
			*((u_int64_t *)currentPkt+1)=rcookie;
	}
	///fix up hdr (80% of the time) when hdr isn't being fuzzed (next payload, version, exchange type, flags, len)
	if( (currentPkt!=NULL) && (index==3) && (location==3) && ((rand()&0xff) > 50) )
	{
		if( !(currentPosition==16) )
			*(currentPkt+16)=0x04;
		if( !(currentPosition==17) )
			*(currentPkt+17)=0x10;
		if( !(currentPosition==18) )
			*(currentPkt+18)=0x02;
		if( !(currentPosition==19) )
			*(currentPkt+19)=0x00;
		if( !(currentPosition>=24 && currentPosition<=27) )
			*((u_int32_t *)currentPkt+6)=htonl(currentPktLen);
	}
	//Might as well fix up the key and nonce hdrs once and a while for fun (60%)
	if( (currentPkt!=NULL) && (index==3) && (location == 3) && ((rand()&0xff) > 100) )
        {	
		if(currentPosition==28)
			*(currentPkt+28)=0x0a;
		if(currentPosition==160)
			*(currentPkt+160)=0x00;
	}
	///we also want to fix up the length if we're fuzzing in key or nonce currentPkt positions (80%)
	if( (currentPkt!=NULL) && (index==3) && (location == 3) && ((rand()&0xff) > 50) )
        {
		if(currentPosition>=32 && currentPosition<=159)
                        *((u_int16_t *)currentPkt+15)=htons(currentPktLen-48);  //180-132=48
                else if(currentPosition>=164 && currentPosition<=179)
                        *((u_int16_t *)currentPkt+81)=htons(currentPktLen-160);  //180-20=160
        }

	///We want to properly use the key & nonce currentPkt, but I need to learn crypto stuff!!!
	

/*5th pkt (recv)*/


/*6th pkt (send)*/
	///Set rcookie & icookie in the case this pkt is the one being fuzzed (% based on previous set)
	if(goodCookie && currentPkt!=NULL)
	{	
		if( (icookieSet==1) && (index==5) && (location == 3) )
			*((u_int64_t *)currentPkt)=icookie;
		if( (index==5) && (location == 3) )
			*((u_int64_t *)currentPkt+1)=rcookie;
	}
	///fix up hdr (80% of the time) when hdr isn't being fuzzed (next payload, version, exchange type, flags, len)
	if( (currentPkt!=NULL) && (index==5) && (location==3) && ((rand()&0xff) > 50) )
	{
		if( !(currentPosition==16) )
			*(currentPkt+16)=0x05;
		if( !(currentPosition==17) )
			*(currentPkt+17)=0x10;
		if( !(currentPosition==18) )
			*(currentPkt+18)=0x02;
		if( !(currentPosition==19) )
			*(currentPkt+19)=0x01;
		if( !(currentPosition>=24 && currentPosition<=27) )
			*((u_int32_t *)currentPkt+6)=htonl(currentPktLen);
	}


/*Moving on to Phase 2*/
/*7th pkt*/

		
	
	return ret;
}
