/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CmdLine is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*
 This is a General Purpose Fuzzing (GPF) suite to fuzz IP networked code.
*/

/*Defines*/
#define RESPECT 1
#define NORESPECT 0
/*Includes*/
#include "headers.h" 
#include "sig.h"
#include "usage.h"
#include "calcRange.h"
#include "readCaptured.h"
#include "random.h"
#include "print.h"

/*Global Variables*/
pcap_dumper_t *dumpfile;
u_char firstTime=0; //0=haven't been through pkt readin loop yet.  1=have been.
u_int loopedOnce=0; //0=no; 1=yes -  To be sure the fuzzer did something

u_int replayMode=0;  //0=normal (client) mode; 1=server mode
u_int printChoice=0; //0=ascii; 1=hex; 2=auto
u_char userFuncName[101];  //holds the name of the userFunc passed from cmdline

ip_address clientIP;
ip_address serverIP;

//It's a little bit ugly have these globals static.
//Up these if you have a maga huge session to replay
u_int seq[3000];  //order in which legs took place  0=client speaking; 1=server speaking
void * Data[1500]; //master list of each of the legs in order
u_int DataSize[1500];  //used to hold the size of each pkt data
u_int Count=0;  //total number of client and server legs
void * readData[1500];  //used to hold a pointer to the data actually read from target
u_int readDataSize[1500];  //used to hold the size of each pkt read from target

int flags = 0; //used to store fd flags

int UDP_SERV=0; //Indicates if we're a UDP server or not; 0=no, 1=yes
int UDP_SERV_READ=0; //Indicates if we've read once with recvfrom to set udpaddr; 0=no, 1=yes
struct sockaddr_in udpaddr;  //added for udp
socklen_t structLen;		//added for udp

struct Tok{
	u_int dataLen;
	u_char * data;
	u_char string;
	u_char numeric;
	u_char bracket;
	u_char quote;
	u_char special;
	u_char other;
	u_int delimLen;
	u_char * delim;
	u_int leg;
	u_char legEnd;
	struct Tok * next;
};	

/*Function prototypes*/
int WRITE(int, u_char *, u_int);
int READ(int, u_int, u_int);
void PRINT(u_int, u_int, u_char *);
int USERFUNC(int, u_int, u_int, u_int, u_char *, u_int, int connfd, struct addrinfo *res);
void FLAGS(int);
void BLOCKING(int);
void NONBLOCKING(int);
void readin(u_char *, const struct pcap_pkthdr *, const u_char *);
int call_readin(u_char *, u_char *);
void txt_readin(u_char *);
void txt_write(int);
void TIME();
int BufferAttack(u_int, u_int, u_int, u_int, int, struct addrinfo *, u_int);
int FormatAttack(u_int, u_int, u_int, u_int, int, struct addrinfo *, u_int);
int LogicAttack(u_int, u_int, u_int, u_int, int, struct addrinfo *, u_int);

u_int Is_String(struct Tok *);
u_int Is_Numeric(struct Tok *);
u_int Is_Bracket(struct Tok *);
u_int Is_Special(struct Tok *);
u_int Is_Quote(struct Tok *);
u_int Is_Other(struct Tok *);
void * Create_Tok();
void * Dup_Tok(struct Tok *);
void Destroy_Tok(struct Tok *);
void Print_Toks(struct Tok *, char *);
int Build_Fuzzy_Leg(u_char **, u_int *, struct Tok *, u_int, u_int);
int Fuzz(struct Tok *);
int Fuzz_Str_Ending(struct Tok *);
int Fuzz_Numeric(struct Tok *);
int Fuzz_Integer(int);
int Fuzz_Parsers(struct Tok *);
u_char * _Fuzz_Parsers(u_char *, u_char);
int Fuzz_BitFlip(struct Tok *);
int Fuzz_Buffer(struct Tok *);
int Fuzz_Format(struct Tok *);

/*UserFuncs go here*/
#include "jabber.h"
#include "db_OtoR.h"
#include "db_RtoA.h"
#include "isa.h"
#include "isakmp.h"
#include "isakmp_serv.h"

/*Main code*/
main(u_int argc, u_char *argv[])
{
	/*Variable Definitions*/
	int listenfd, ERROR, fp_src, fp_dst;
	u_int in, none;
	u_short len;
	struct hostent *host_info;	//added for udp
	struct sockaddr_in cliaddr;	//added for udp
	u_char clihost[16];		//added for udp  -- probably won't need except for printing
	u_char *packet;			//for -R
	u_int i, j, k, l, x, y;
	int c;
	u_int loops=0;

	int connfd;
	struct addrinfo hints, *res;
	u_int userFunc=1; //0=no user func; 1=yes -- to increasing GPF's speed
	u_int cycleMode=1; //0=close fd after attack; 1=keep replaying if possible
	u_int readWait = 3;
	
	u_char *captureFilename;
	u_int type;
	u_char *host, *port;
	u_int delay=100;

/*Setup signal handling stuff*/
signal(SIGINT, Sig);
signal(SIGQUIT, Quit);
signal(SIGPIPE, Pipe);
 /*Master switch - Determine the mode we will run in*/
 c = getopt(argc, (char * const)argv, "R:CG:Ph:");
 switch (c)
 {
   case 'R':  /*Pure Fuzz*/
	/*check number of command line arguments*/
	if (argc < 6)
		random_usage();
	/*Handle 2nd and 3rd arg: put into vars*/
	host = argv[1];
	port = argv[2];
	/*Handle 4th arg: perform getaddrinfo() call*/
	bzero(&hints, sizeof(hints));
	if ( strcasecmp(argv[3], "UDP") == 0)
		hints.ai_socktype = SOCK_DGRAM;
	else if ( strcasecmp(argv[3], "TCP") == 0 )
		hints.ai_socktype = SOCK_STREAM;
	else
		random_usage();
	/*Handle 5th arg: Seed random number generator*/	
	srand( (u_int)(atoi(argv[4])) );
	/*Handle 6th arg: Input Delay*/	
	delay = (u_int)(atoi(argv[5]));
	/*Prepare Networking info*/
	if( (ERROR = getaddrinfo(host, port, &hints, &res)) !=0) {
		TIME(); fprintf(stderr, "{{{%s}}}\n", gai_strerror(ERROR)); exit(-1); }
	/*Main code to spew random pkts*/
	j=0;
	in=0;
	calcRange( (u_int *)&none, (u_int *)&x, (u_int *)&y, optarg);
	while(j<=y || none==1)
	{
		if( (connfd = setup_client (res)) == SOCKET_ERROR) {
			TIME(); fprintf (stderr, "{{{Setup_client failed!}}}\n\n"); }
		len=RAND(15000);
		packet=malloc(len);
		for(i=0;i<len;i++)
		{
			packet[i]=RAND(256);
		}
		if( (j>=x) && (j<=y) || (none == 1) )
		{
			printf("Pkt: %d  len=%6.0d\t", j, len);
			write(connfd, packet, len);
			printf("- sent data\n", j);
			usleep(delay);
		}
		free(packet);
		j++;
		close(connfd);
	}
	printf("\n\n Successful Finish.  Did we crash your target?\n\n");
	exit(0);
  
  case 'C': /*Convert mode - allows you to convert a libpcap file to a text file*/
	/* check args.  */
        if (argc < 3) {
       		TIME(); fprintf(stderr, "{{{Not enough args}}}\n"); exit(-1); } 

	/*open files with error check*/
        umask(0);
        if( (fp_dst = open(argv[2], O_WRONLY|O_CREAT, 0666)) <0) {
                TIME(); fprintf(stderr, "{{{Couldn't open dst_file.}}}\n"); exit(-1); }
	/*Calls readin which reads ethereal dump into the global array*/
	if ( call_readin(argv[1], "dumpout") !=0) {
		TIME(); fprintf(stderr, "{{{Problem reading in capture file!}}}\n\n"); exit(-1); }

	/*write global data out to txt.gpf type file*/
	txt_write(fp_dst);

	/*close file descriptor*/
        if( (close(fp_dst) <0)) {
                perror("close fp_dst"); exit(-1); }
	printf("\n\n\tDone.\n\n");
	exit(0);

   case 'G': /*GPF*/
	/*check number of command line args*/
	if (argc < 23){
		TIME(); fprintf(stderr, "{{{Not enough args}}}\n"); GPF_usage();}
	if ( optarg!=NULL && (strcmp("h", optarg))==0 )
		GPF_usage();
	c=*optarg;
	/*declare more vars for main()*/
	u_int startLeg=0;
	u_int stopLeg=0;
	u_int currentLeg=0;
	u_int _currentLeg=0;
	u_int tempLeg=0;
	u_int startPosition=0;
	u_int stopPosition=-1;
	u_int _stopPosition=1;
	u_int currentPosition=0;
	u_int _currentPosition=0;
	u_int startSize=0;
	u_int incSize=255;
	u_int stopSize=17000;
	u_int attackSize=0;
	u_int attackChar=0x41;
	u_int extraAttacks=0; //default is one attack (on current line) per cycle
	u_int _extraAttacks=0;
	u_int attacked=0;
	/*handle args*/
	captureFilename = argv[1];
	type = (u_int)(atoi(argv[2]));
	/*Call the appropriate readin function*/
	if (type ==0)
	{
		txt_readin(captureFilename);	

	}
	else if (type ==1)
	{
		/*Calls readin which reads ethereal dump into the global array*/
		if ( call_readin(captureFilename, "dumpout") !=0) {
			TIME(); fprintf(stderr, "{{{Problem reading in capture file!}}}\n\n"); exit(-1); }
	}
	else{
		TIME(); fprintf(stderr, "{{{Bad type}}}\n"); GPF_usage();}
	/*reserve spaced for the readData*/
	for(x=0; x<Count; x++)
	{
		readData[x]=malloc(1500);	
		readDataSize[x]=0;	
	}
	/*handle the rest of the args*/
	replayMode = (u_int)(atoi(argv[3]));
	host = argv[4];
	port = argv[5];
	bzero(&hints, sizeof(hints));
	if ( strcasecmp(argv[6], "UDP") == 0)
		hints.ai_socktype = SOCK_DGRAM;
	else if ( strcasecmp(argv[6], "TCP") == 0 )
		hints.ai_socktype = SOCK_STREAM;
	else
		GPF_usage();
	srand( (u_int)(atoi(argv[7])) );
	delay = (u_int)(atoi(argv[8]));
	startLeg = (u_int)(atoi(argv[9]));
	if( (strcmp("+", argv[10]))==0)
		stopLeg=Count;
	else
		stopLeg = (u_int)(atoi(argv[10]));
	startPosition = (u_int)(atoi(argv[11]));
	if( (strcmp("+", argv[12]))==0)
		stopPosition=-1;
	else
		stopPosition = (u_int)(atoi(argv[12]));
	startSize = (u_int)(atoi(argv[13]));
	incSize = (u_int)(atoi(argv[14]));
	stopSize = (u_int)(atoi(argv[15]));
	sscanf(argv[16],"%x", &attackChar);
	if( (printChoice = (u_int)(atoi(argv[17]))) > 2)
	{	
		fprintf(stderr, "{{{Bad print choice: needs to be 0, 1, or 2}}}\n\n");
		GPF_usage();
	}
	if( (strcmp("none", argv[18]))==0)
		userFunc=0;
	else
		strncpy(userFuncName, argv[18], 100);
	if( (cycleMode = (u_int)(atoi(argv[19]))) > 2)
	{	
		fprintf(stderr, "{{{Bad cycle mode: needs to be 0 or 1}}}\n\n");
		GPF_usage();
	}
	readWait = (u_int)(atoi(argv[20]));
	if( (extraAttacks = (u_int)(atoi(argv[21]))) < 0)
	{	
		fprintf(stderr, "{{{Bad extraAttacks: needs to be 0 or more}}}\n\n");
		GPF_usage();
	}
	/*Prepare Networking info*/
	if( (ERROR = getaddrinfo(host, port, &hints, &res)) !=0) {
		TIME(); fprintf(stderr, "{{{%s}}}\n", gai_strerror(ERROR)); exit(-1); }
	/*Bind to server IP for TCP */
	if( replayMode == 1 && hints.ai_socktype == SOCK_STREAM)
	{
		if( (listenfd = setup_server(res)) == SOCKET_ERROR) {
			TIME(); fprintf(stderr, "{{{Setup_server failed}}}\n\n"); exit(-1); }
	}
	/*for the very ugly case that we're a UDP server*/
	if( replayMode == 1 && hints.ai_socktype == SOCK_DGRAM)
	{
		UDP_SERV=1;
		structLen=sizeof(udpaddr);
		bzero(&udpaddr, structLen);
		/*udpaddr.sin_family=AF_INET;
		if  ( (host_info = gethostbyname(host)) == NULL)
		{
			if (inet_pton(AF_INET, host, &udpaddr.sin_addr) <= 0 ){
				perror("inet_pton"); exit(-1); }
		}
		else
		{
			udpaddr.sin_addr.s_addr = (((struct in_addr *) host_info->h_addr_list[0])->s_addr);
		}
		udpaddr.sin_port = htons(atoi(port));
		*/
	}
	printf("%d total legs\n\n", Count);
	printf("Beginning replay attack:\n");
        
	/*fuzz each session according to user specified parameters*/
	///Outer loop for switching legs to fuzz
	for(currentLeg=startLeg; currentLeg<stopLeg; currentLeg++)
	{
	     if( (replayMode==0 && seq[currentLeg]==0) || (replayMode==1 && seq[currentLeg]==1))
	     {
		///Inner loop 1 for increasing the size of attack
		for(attackSize=startSize; attackSize<stopSize; attackSize=attackSize+incSize)
		{	
			if( (stopPosition==-1) || (stopPosition>DataSize[currentLeg]) ) 
				_stopPosition=DataSize[currentLeg];	
			else
				_stopPosition=stopPosition;	
			///Inner loop 2 for moving along the line (position) while using same attackSize
			for(currentPosition=startPosition; currentPosition<=_stopPosition; currentPosition++)
 			{
				///Connect to target if client mode
				if( replayMode == 0)
				{
					if( (connfd = setup_client (res)) == SOCKET_ERROR) {
						TIME(); fprintf(stderr, "{{{Setup_client failed}}}\n\n"); exit(-1); }
				}
				///Accept incoming connections if TCP server mode
				if( replayMode==1 && hints.ai_socktype==SOCK_STREAM)
				{
					if( (connfd = accept(listenfd, (struct sockaddr *)NULL, NULL)) < 0) {
						TIME(); fprintf(stderr, "{{{server setup failed}}}\n\n"); exit(-1); }
				}
				///Bind to server IP for UDP
				if( replayMode == 1 && hints.ai_socktype == SOCK_DGRAM)
				{
					if( (connfd = setup_server(res)) == SOCKET_ERROR) {
						TIME(); fprintf(stderr, "{{{Setup_server failed}}}\n\n"); exit(-1); }
				}
				///GET fd flags
				FLAGS(connfd);
				///Inner loop 3.  Send, Read, Process, and/or Attack.
				_extraAttacks=extraAttacks;
				_currentLeg=currentLeg;
				tempLeg=0;
				while(tempLeg<Count)
				{
					ERROR=0;
					usleep(delay);	
					if( (replayMode==0 && seq[tempLeg]==0) || (replayMode==1 && seq[tempLeg]==1) )
					{
						if(tempLeg!=_currentLeg)
						{
							WRITEAGAIN:
							if(userFunc)
								USERFUNC(0, tempLeg, 0, currentPosition, Data[tempLeg],
												 DataSize[tempLeg], connfd, res);
							printf("[%d][-->]  ", tempLeg);
							PRINT(DataSize[tempLeg], printChoice, Data[tempLeg]);
								printf("\n");
							if(WRITE(connfd, Data[tempLeg], DataSize[tempLeg]) < 1)
							{
								TIME(); perror("{{Dead Socket (W <1)}}"); break; 
							}
							tempLeg++;
						}
						else
						{
						     switch(c)
						     {
       							/*Replay verification with no attacks inserted*/
						       case 'r':
								cycleMode=1;
								currentLeg=stopLeg;
								attackSize=stopSize;
								currentPosition=_stopPosition+1;
								goto WRITEAGAIN;
							break;
							/*GPF Buffer overflow attack mode*/
						       case 'b':
							   attacked=1;
							   if(currentPosition<=DataSize[tempLeg])
							   {
								if ( (ERROR=BufferAttack(_currentLeg, currentPosition,
									attackSize, attackChar, connfd, res, userFunc)) == 1)
									break;
								tempLeg++;
								if(_extraAttacks!=0)
								{
									_extraAttacks--;
									_currentLeg++;
									for(i=tempLeg; !((replayMode==0 && seq[i]==0) ||
										(replayMode==1 && seq[i]==1)); i++ )
										_currentLeg++;
								}	
							    }
							    else
							    {
								goto WRITEAGAIN;
							    }
							break;
							/*Format string GPF attack mode*/
						       case 'f':	
							   attacked=1;
							   if(currentPosition<=DataSize[tempLeg])
							   {
								if ( (ERROR=FormatAttack(_currentLeg, currentPosition,
									attackSize, attackChar, connfd, res, userFunc)) == 1)
									break;
								tempLeg++;
								if(_extraAttacks!=0)
								{
									_extraAttacks--;
									_currentLeg++;
									for(i=tempLeg; !((replayMode==0 && seq[i]==0) ||
										(replayMode==1 && seq[i]==1)); i++ )
										_currentLeg++;
								}	
							    }
							    else
							    {
								goto WRITEAGAIN;
							    }
							break;
							/*Logic GPF attack mode*/
						       case 'l':
							   attacked=1;
							   if(currentPosition<=DataSize[tempLeg])
							   {
								if ( (ERROR=LogicAttack(_currentLeg, currentPosition,
									attackSize, attackChar, connfd, res, userFunc)) == 1)
									break;
								tempLeg++;
								if(_extraAttacks!=0)
								{
									_extraAttacks--;
									_currentLeg++;
									for(i=tempLeg; !((replayMode==0 && seq[i]==0) ||
										(replayMode==1 && seq[i]==1)); i++ )
										_currentLeg++;
								}	
							    }
							    else
							    {
								goto WRITEAGAIN;
							    }
							break;
						       default:
								fprintf(stderr, "{{{Bad attack mode: use r, b, f, l}}}\n\n");
								GPF_usage();
			   			  }
						}
					    	if (attacked==1)
						{
							attacked=0;
							if(ERROR)	
								break;	
							if( (cycleMode==0) && (_extraAttacks==0) )
								break;
						}
					}
					if( (replayMode==0 && seq[tempLeg]==1) || (replayMode==1 && seq[tempLeg]==0) )
					{
						READAGAIN:
						if( (x=READ(connfd, tempLeg, readWait)) == -2) {
							perror("{{Couldn't Read Socket (R -2)}}"); }
						else if(x == -1) {
							TIME(); perror("{{Dead Socket (R -1)}}"); break; }
						tempLeg++;
						if(userFunc)
						{
							y=USERFUNC(x, tempLeg, 1, currentPosition, readData[tempLeg],
											readDataSize[tempLeg], connfd, res);
							if( y==-1)
							{
								printf("{{Didn't like what we got...Let's read again.}}\n");
								goto READAGAIN;
							}
							if( y==1)
							{
								printf("{{Didn't like what we got...Let's write again.}}\n");
								tempLeg--;
								goto WRITEAGAIN;
							}
						}
					}
				}
				///Clean up and head to the bottom of this big o' loop	
				printf("=================================Total Loops: %d====================================\n\n", loops);
				loops++;
				if(userFunc)
					USERFUNC(0, _currentLeg, 4, currentPosition, 0, 0, connfd, res);
				usleep(delay);
				close(connfd);
				if(userFunc)
					USERFUNC(0, _currentLeg, 5, currentPosition, 0, 0, connfd, res);
				///Checks to be sure we went around at least once
				if(!loopedOnce)
					loopedOnce=1;
			}
		}
	    }
	}
	break;

   case 'P': /*Pattern GPF*/
	/*check number of command line arguments and handle them*/
	if (argc < 15){
		TIME(); fprintf(stderr, "{{{Not enough args}}}\n"); Pattern_GPF_usage();}
	captureFilename = argv[1];
	type = (u_int)(atoi(argv[2]));
	/*Call the appropriate readin function*/
	if (type ==0)
	{
		txt_readin(captureFilename);	

	}
	else if (type ==1)
	{
		/*Calls readin which reads ethereal dump into the global array*/
		if ( call_readin(captureFilename, "dumpout") !=0) {
			TIME(); fprintf(stderr, "{{{Problem reading in capture file!}}}\n\n"); exit(-1); }
	}
	else
	{
		TIME(); fprintf(stderr, "{{{Bad type}}}\n"); GPF_usage();
	}
	/*reserve spaced for the readData*/
	for(x=0; x<Count; x++)
	{
		readData[x]=malloc(1500);	
		readDataSize[x]=0;	
	}
	/*handle the rest of the args*/
	replayMode = (u_int)(atoi(argv[3]));
	host = argv[4];
	port = argv[5];
	bzero(&hints, sizeof(hints));
	if ( strcasecmp(argv[6], "UDP") == 0)
		hints.ai_socktype = SOCK_DGRAM;
	else if ( strcasecmp(argv[6], "TCP") == 0 )
		hints.ai_socktype = SOCK_STREAM;
	else
		Pattern_GPF_usage();
	srand( (u_int)(atoi(argv[7])) );
	delay = (u_int)(atoi(argv[8]));
	if( (printChoice = (u_int)(atoi(argv[9]))) > 2)
	{	
		fprintf(stderr, "{{{Bad print choice: needs to be 0, 1, or 2}}}\n\n");
		Pattern_GPF_usage();
	}
	if( (strcmp("none", argv[10]))==0)
		userFunc=0;
	else
		strncpy(userFuncName, argv[10], 100);
	readWait = (u_int)(atoi(argv[11]));
	u_char * delim; 
	delim=argv[12];
	u_int delimLen;
	delimLen = (u_int)(atoi(argv[13]));
	/*Prepare Networking info*/
	if( (ERROR = getaddrinfo(host, port, &hints, &res)) !=0) {
		TIME(); fprintf(stderr, "{{{%s}}}\n", gai_strerror(ERROR)); exit(-1); }
	/*Bind to server IP for TCP */
	if( replayMode == 1 && hints.ai_socktype == SOCK_STREAM)
	{
		if( (listenfd = setup_server(res)) == SOCKET_ERROR) {
			TIME(); fprintf(stderr, "{{{Setup_server failed}}}\n\n"); exit(-1); }
	}
	/*for the very ugly case that we're a UDP server*/
	if( replayMode == 1 && hints.ai_socktype == SOCK_DGRAM)
	{
		UDP_SERV=1;
		structLen=sizeof(udpaddr);
		bzero(&udpaddr, structLen);
		/*udpaddr.sin_family=AF_INET;
		if  ( (host_info = gethostbyname(host)) == NULL)
		{
			if (inet_pton(AF_INET, host, &udpaddr.sin_addr) <= 0 ){
				perror("inet_pton"); exit(-1); }
		}
		else
		{
			udpaddr.sin_addr.s_addr = (((struct in_addr *) host_info->h_addr_list[0])->s_addr);
		}
		udpaddr.sin_port = htons(atoi(port));
		*/
	}
	/*Tokenize*/
	//add code to check if delim or delimLen are 0 or NULL.
	//if they are, we should look for size:data pattern in binary data
	//thus there will be no delims between tokens.
	//I must remember this when coding the fuzz tokens loop
	struct Tok * Toks;
	struct Tok * tokptr;
	u_char * data=NULL;
	u_int size;
	u_char * tok=NULL;
	u_char * temp=NULL;
	u_int tokSize=0;
	u_int leg=0;

	leg=0;
	i=0;
	while(leg<Count)
	{
		if( (replayMode==0 && seq[leg]==0) || (replayMode==1 && seq[leg]==1) )
		{
			data=Data[leg];
			size=DataSize[leg];
			u_int totalTokSize=0;
			while(1)
			{
				//create tok; i keeps inc'ing but is never used
				if( i == 0)
				{
					Toks=Create_Tok();
					tokptr=Toks;
				}
				else
				{
					tokptr->next=Create_Tok();
					tokptr=tokptr->next;	
				}
				//find next tok and tokSize
				temp=memmem(data, (size-totalTokSize), delim, delimLen);
				if (temp == NULL)
				{
					tokSize=size-totalTokSize;
					tokptr->legEnd=1;
				}
				else
					tokSize=temp-data;
				//make room and copy data to tok
				if(tokSize < 0)
				{
						fprintf(stderr, "{{{Tok0: Bad tokSize!!}}}\n"); exit(-1);
				}
				if(tokSize == 0)
				{
					tok=malloc(delimLen);
					memcpy(tok, delim, delimLen); 
					tokptr->dataLen=delimLen;
				}
				else
				{
					if((tok=malloc(tokSize)) == NULL)
					{
						fprintf(stderr, "{{{Tok1: No memory left!!}}}\n"); exit(-1);
					}
					memcpy(tok, data, tokSize); 
					tokptr->dataLen=tokSize;
				}
				data=data+tokSize+delimLen;
				totalTokSize=totalTokSize+tokSize+delimLen;
				tokptr->data=tok;
				tokptr->leg=leg;
				//is token a string?
				if(Is_String(tokptr))
				{
					tokptr->string=1;
					//If so, does the string contain numbers?
					if(Is_Numeric(tokptr))
						tokptr->numeric=1;
					//does it contain bracket?
					if(Is_Bracket(tokptr))
						tokptr->bracket=1;
					//does it contain quote?
					if(Is_Quote(tokptr))
						tokptr->quote=1;
					//does it contain special?
					if(Is_Special(tokptr))
						tokptr->special=1;
					//does it others
					if(Is_Other(tokptr))
						tokptr->other=1;
				}
				else //handle binary protocols
				{
				}
				//we inc only for first time check?
				i++;
				//if leg end bail out
				if(tokptr->legEnd == 1)
					break;
				//else add the delimeter info to token
				else
				{
					tokptr->delimLen=delimLen;
					if((tokptr->delim=malloc(tokptr->delimLen)) == NULL)
					{
						fprintf(stderr, "{{{Tok2: No memory left!!}}}\n"); exit(-1);
					}
					memcpy(tokptr->delim, delim, tokptr->delimLen);
				}
			//For testing, comment out when not using
			//Print_Toks(Toks, "all");
			//printf("------------loop=%d, leg=%d-------------\n", i, leg);
			}
		}
		leg++;
	}
	//for testing...print final linked list of tokens
	//Print_Toks(Toks, "all");

	/*Pattern fuzz*/
	printf("%d total legs\n\n", Count);
	printf("Beginning replay attack:\n");
	while(1)
	{
		//Connect to target if client mode
		if( replayMode == 0)
		{
			if( (connfd = setup_client (res)) == SOCKET_ERROR) {
				TIME(); fprintf(stderr, "{{{Setup_client failed}}}\n\n"); exit(-1); }
		}
		//Accept incoming connections if TCP server mode
		if( replayMode==1 && hints.ai_socktype==SOCK_STREAM)
		{
			if( (connfd = accept(listenfd, (struct sockaddr *)NULL, NULL)) < 0) {
				TIME(); fprintf(stderr, "{{{server setup failed}}}\n\n"); exit(-1); }
		}
		//Bind to server IP for UDP
		if( replayMode == 1 && hints.ai_socktype == SOCK_DGRAM)
		{
			if( (connfd = setup_server(res)) == SOCKET_ERROR) {
				TIME(); fprintf(stderr, "{{{Setup_server failed}}}\n\n"); exit(-1); }
		}
		//GET fd flags
		FLAGS(connfd);
		//Main fuzzy loop
		leg=0;
		//respect
		if( RAND(100) < 90 )
		{
			while(leg<Count)
			{
				if( (replayMode==0 && seq[leg]==0) || (replayMode==1 && seq[leg]==1) )
				{
					usleep(delay);	
					u_char * Packet=NULL;
					u_int totalLen=0;
					P_WRITEAGAIN:
					//Build Fuzzy Leg
					if( (Build_Fuzzy_Leg(&Packet, &totalLen, Toks, RESPECT, leg)) == -1 )
					{
						TIME(); fprintf(stderr, "{{Build_Fuzzy_Leg failed!}}\n"); exit(-1);
					}
					//process attack??
					if(userFunc)
						USERFUNC(0, leg, 1, 0, Packet, totalLen, connfd, res);
					//print attack
					printf("[%d][-p:%d->] ",leg, totalLen);
					if(totalLen<150)
					{
						PRINT(totalLen, printChoice, Packet);
							printf("\n");
					}
					else
					{
						PRINT(150, printChoice, Packet);
							printf("...TRUNCATED...\n");
					}
					//send attack
					if(WRITE(connfd, Packet, totalLen) < 1)
					{
						TIME(); perror("{{{Pattern attack write error}}}");
					}
					free(Packet);
					leg++;
				}	
				else if( (replayMode==0 && seq[leg]==1) || (replayMode==1 && seq[leg]==0) )
				{
					usleep(delay);	
					P_READAGAIN:
					if( (x=READ(connfd, leg, readWait)) == -2) {
						perror("{{Couldn't Read Socket (R -2)}}"); }
					else if(x == -1) {
						TIME(); perror("{{Dead Socket (R -1)}}"); break; }
					leg++;
					if(userFunc)
					{
						y=USERFUNC(x, leg, 2, 0, readData[leg], readDataSize[leg], connfd, res);
						if( y==-1)
						{
							fprintf(stderr, "{{Didn't like what we got...Let's read again.}}\n");
							goto P_READAGAIN;
						}
						if( y==1)
						{
							fprintf(stderr, "{{Didn't like what we got...Let's write again.}}\n");
							leg--;
							goto P_WRITEAGAIN;
						}
					}
				}
				else
				{
					TIME();fprintf(stderr, "{{{Bad replayMode or seq[]}}}\n");exit(-1);
				}
			}
		}
		//NO respect
		else
		{
			while(leg<Count)
			{
				if( RAND(100) < 65 )
				{
					usleep(delay);	
					u_char * Packet=NULL;
					u_int totalLen=0;
					P_WRITEAGAIN2:
					//Build Fuzzy Leg
					if( (Build_Fuzzy_Leg(&Packet, &totalLen, Toks, NORESPECT, leg)) == -1)
					{
						TIME(); fprintf(stderr, "{{Build_Fuzzy_Leg failed!}}\n"); exit(-1);
					}
					//process attack??
					if(userFunc)
						USERFUNC(0, leg, 1, 0, Packet, totalLen, connfd, res);
					//print attack
					printf("[%d][-p:%d->] ",leg, totalLen);
					if(totalLen<150)
					{
						PRINT(totalLen, printChoice, Packet);
							printf("\n");
					}
					else
					{
						PRINT(150, printChoice, Packet);
							printf("...TRUNCATED...\n");
					}
					//send attack
					if(WRITE(connfd, Packet, totalLen) < 1)
					{
						TIME(); perror("{{{Pattern attack write error}}}");
					}
					free(Packet);
					leg++;
				}	
				else
				{
					usleep(delay);	
					P_READAGAIN2:
					if( (x=READ(connfd, leg, readWait)) == -2) {
						perror("{{Couldn't Read Socket (R -2)}}"); }
					else if(x == -1) {
						TIME(); perror("{{Dead Socket (R -1)}}"); break; }
					leg++;
					if(userFunc)
					{
						y=USERFUNC(x, leg, 2, 0, readData[leg], readDataSize[leg], connfd, res);
						if( y==-1)
						{
							fprintf(stderr, "{{Didn't like what we got...Let's read again.}}\n");
							goto P_READAGAIN2;
						}
						if( y==1)
						{
							fprintf(stderr, "{{Didn't like what we got...Let's write again.}}\n");
							leg--;
							goto P_WRITEAGAIN2;
						}
					}
				}
			}
		}
		//loop	
		printf("=================================Total Loops: %d====================================\n\n", loops);
		loops++;
		close(connfd);
		//checks to be sure we went around at least once
		if(!loopedOnce)
			loopedOnce=1;
	}
	break; //should never reach this

   case 'h': /*Print Usage*/
		if ( optarg!=NULL && (strcmp("R", optarg))==0 )
			random_usage();
		else if ( optarg!=NULL && (strcmp("C", optarg))==0 )
			doctor_usage();
		else if ( optarg!=NULL && (strcmp("G", optarg))==0 )
			GPF_usage();
		else if ( optarg!=NULL && (strcmp("P", optarg))==0 )
			Pattern_GPF_usage();
		else
			main_usage();

   default:  /*The default switch*/
        fprintf(stderr, "{{{Bad args -- Did you choose a mode?}}}\n\n");
        main_usage();
 }
 //from fall out of -G	  
 if(loopedOnce)
    printf("\n\n Successful Finish.  Did we crash your target?\n\n");
 else {
    TIME(); fprintf(stderr,"{{{Didn't loop even once. Check command line params or capture file.}}}\n\n"); }
 exit(0);
}

/*send data out, handle UDP as well*/
int WRITE(int fd, u_char * data, u_int size)
{
	int i, x; 

	NONBLOCKING(fd);
	for(i=0; i<2; i++)
	{
		if(UDP_SERV)
			x=sendto(fd, data, size, 0, (struct sockaddr *) &udpaddr, structLen);
		else
			x=write(fd, data, size);
		if (x == -1 || x == 0 )
			return x;
		else if (x == size)
			return x;
		printf("~~waiting to write data~~\n");
		sleep(1);
	}
	printf("~~~~Failed to write, moving on. %d sent.~~~~\n", x);
	return x;
}

/* Read data off the wire and print it; compares to stored. */
int READ(int fd, u_int index, u_int wait)
{
	u_int equal=0; //0 means the received and stored were the same
	int x=0;
	int y=0;
	int retval=0;

	if(UDP_SERV && !UDP_SERV_READ)
	{
		x=recvfrom(fd, readData[index], 1500, 0, (struct sockaddr *) &udpaddr, &structLen);
		UDP_SERV_READ=1;
	}
	else
	{
		NONBLOCKING(fd);
		while( (x=read(fd, readData[index], 1500)) != 0)
		{
			y++;
			if(x > 0)
				break;
			usleep(1);
			if ( y > (13+(wait)) )
			{
				printf("~~~~moving on without that data~~~~\n");
				fflush(stdout);
				break;
			}
			if (y>13)
			{
				printf("~~waiting to read data~~\n");
				sleep(1);
			}
		}
	}
	if( (x < 1) || (x > 1500) )
	{
		if(errno == EAGAIN)
			retval=-2;  ///Server didn't send data, but socket is still ok
		else
			retval=-1;
	}
	else
	{	
		readDataSize[index]=x;
		if(Data[index] !=0)
		{
			equal=memcmp(Data[index], readData[index], readDataSize[index]);
			if(equal==0)
				printf("  [<--][%d] ", index);
			else
				printf("  [<-DIFF-][%d] ", index);
		}
		else
			printf("  [<-NoDataCMP-]\n");
		PRINT(readDataSize[index], printChoice, readData[index]);
			printf("\n");
	}
	return retval;
}

/*Calls userFunc if applicable*/
/*index is the leg of protocol*/
/*location indicates one of three spots that USERFUNC is called
0=before pre_write_loop, 1=after pre_read_loop, 2=pre_attack construction, 3=pre_send attack, 4=pre_close, 5=after close*/
/*Currently there's only one place where the return value from USERFUNC is checked: Location==1 (after the read) */
int USERFUNC(int readRet, u_int index, u_int location, u_int currentPosition, u_char * currentPkt, u_int currentPktLen, int connfd, struct addrinfo *res)
{
	int ret;

	if( (strcmp(userFuncName, "jabber")) == 0)
		jabber(index, location, connfd, res);
	else if( (strcmp(userFuncName, "db_OtoR")) == 0)
		db_OtoR(index, location, connfd, res);
	else if( (strcmp(userFuncName, "db_RtoA")) == 0)
		db_RtoA(index, location, connfd, res);
	else if( (strcmp(userFuncName, "isakmp")) == 0)
		ret=isakmp(readRet, index, location, currentPosition, currentPkt, currentPktLen);
	else if( (strcmp(userFuncName, "isakmp_serv")) == 0)
		ret=isakmp_serv(readRet, index, location, currentPosition, currentPkt, currentPktLen);
	else { fprintf(stderr, "{{{Bad userFunc}}}\n"); exit(-1); }
	
	return ret;
}

/*Function to globally store flags*/
void FLAGS(int fd)
{
	if ((flags = fcntl(fd, F_GETFL)) == -1) {
		TIME(); fprintf(stderr, "{{{fcntl(F_GETFL)}}}"); exit(-1); }
	
}
/*Function to change a file discriptor to blocking*/
void BLOCKING(int fd)
{
	if (fcntl(fd, F_SETFL, flags) == -1) {
		TIME(); fprintf(stderr, "{{{fcntl(F_SETL, 0_NONBLOCK)}}}\n"); exit(-1); }
}
/*Function to change a file discriptor to non-blocking*/
void NONBLOCKING(int fd)
{
	if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1) {
		TIME(); fprintf(stderr, "{{{fcntl(F_SETL, 0_NONBLOCK)}}}\n"); exit(-1); }
}

/*Read the txt.gpf data and sizes in the global arrays*/
void txt_readin(u_char * file)
{
	/*local vars*/	
	int fp_src;
	int fsize=0;
	int x=0;
	int y=0;
	int block=256;
	u_char blockptr[block];
	u_char * fileData=NULL;
	u_char * ptr=NULL;

	/*open files with error check*/
        umask(0);
        if( (fp_src = open(file, O_RDONLY|O_CREAT, 0666)) <0) {
                TIME(); fprintf(stderr, "{{{Couldn't open src_file.}}}\n"); exit(-1); }
	/*read data into array*/
	while ( (x=read(fp_src, blockptr, block)) >0 )
	{
		if(x>0)
			fsize=fsize+x;
		fileData=realloc(fileData, fsize);
		memcpy(fileData+y, blockptr, x);
		if(x>0)
			y=y+x;
	}
	if(fsize == 0){
		TIME(); fprintf(stderr, "{{{NULL FILE - txt_readin}}}\n"); exit(-1);}
	/*parse data and store in global arrays*/
	ptr=fileData;
	while ( 1 )	
	{
		if( (ptr=strstr(ptr, "Source:")) !=NULL )
		{
			if ( *(ptr+7)=='C')
			{
				ptr=strstr(ptr, "Size:");
				DataSize[Count]= (atoi(ptr+5));	
				ptr=strstr(ptr, "Data:");
				if( DataSize[Count] > 0)
				{
					Data[Count]=malloc(DataSize[Count]);
               				memcpy(Data[Count], ptr+5, DataSize[Count]);
					ptr=ptr+5+DataSize[Count];
					seq[Count]=0;
        				Count++;
				}
			}
			else if ( *(ptr+7)=='S')
			{
				ptr=strstr(ptr, "Size:");
				DataSize[Count]= (atoi(ptr+5));	
				ptr=strstr(ptr, "Data:");
				if( DataSize[Count] > 0)
				{
					Data[Count]=malloc(DataSize[Count]);
               				memcpy(Data[Count], ptr+5, DataSize[Count]);
					ptr=ptr+5+DataSize[Count];
					seq[Count]=1;
        				Count++;
				}
			}
			else{
				TIME(); fprintf(stderr, "{{{Bad Source - txt_readin: Count=%d}}}\n", Count); exit(-1);}
		}
		else
			break;
	}
	/*close file descriptor*/
        if( (close(fp_src) <0)) {
                perror("close fp_src"); exit(-1); }
}

/*Write global array out to txt.gpf style file*/
void txt_write(int fd)
{
	/*local vars*/	
	int i, lineSize;
	u_char buf[1600];
	
	/*parse data and store in global arrays*/
	for(i=0; i<Count; i++)	
	{
		memset(buf, 0x00, 1600);
		if(seq[i])
		{
			snprintf(buf,1599, "Source:S Size:%.4d Data:", DataSize[i]);
			lineSize=(strlen(buf) + DataSize[i]);
			memcpy(buf+(strlen(buf)), Data[i], DataSize[i]);
		}
		else
		{
			snprintf(buf,1599, "Source:C Size:%.4d Data:", DataSize[i]);
			lineSize=(strlen(buf) + DataSize[i]);
			memcpy(buf+(strlen(buf)), Data[i], DataSize[i]);
		}
		if ( write(fd, buf, lineSize) < 1){
			TIME(); perror("{{{write proble in txt_write()}}}"); exit(-1);}
	}
	printf("\nWriting %d lines to file...\n", i);
}

/*Reads in and copies pkt data into global array*/
void readin(u_char *temp1, const struct pcap_pkthdr *header, const u_char *pkt_data)
{
	u_int layer2Len=14;  ///14 bytes is the length of ethernet
        u_int i, IPlen, TCPlen, UDPlen, dataLen;
        ip_header *IPheader;
        tcp_header *TCPheader;
        udp_header *UDPheader;
        u_char * data;

	/*This is a special pcap call that I didn't really look into, but it works. :) */
        pcap_dump((u_char *) dumpfile, header, pkt_data);

        /* Get the position of the ip header */
        IPheader = (ip_header *)(pkt_data + layer2Len);
        IPlen = (IPheader->ver_ihl & 0xf) * 4;

	/*Handle TCP*/
        if(IPheader->protocol==0x06)
	{
		///Set TCPheader and calculate data length
                TCPheader = (tcp_header *) ((u_char*)IPheader + IPlen);
                TCPlen= ((TCPheader->hlen & 0xf0) >> 4) * 4;
                dataLen=( (header->len)-layer2Len-IPlen-TCPlen);

		///For the first time through we set the client & server IP
		///The IP to send the SYN will be considered the client
		if ( (firstTime == 0) && (TCPheader->flags==0x02) )
		{
			firstTime = 1;
			clientIP = IPheader->saddr;
			serverIP = IPheader->daddr;
			printf("Found in Capture file:\nclientIP=%d.%d.%d.%d dport=%d --> ",
				clientIP.byte1,clientIP.byte2,clientIP.byte3,clientIP.byte4,ntohs(TCPheader->dport));
			printf("serverIP=%d.%d.%d.%d\n",serverIP.byte1,serverIP.byte2,serverIP.byte3,serverIP.byte4);
			fflush(stdout);
		}
		
		///If there's data (plus check if PSK/ACK) put it in appropriate array
		///if ( (dataLen>0) && (TCPheader->flags==0x18) )
		///Since data can also be in just ACK's had to change to
		if ( (dataLen>0) && (memcmp((u_char *)(pkt_data+layer2Len+IPlen+TCPlen), "\x00\x00\x00\x00\x00\x00", 6) !=0) )
		{
			if(IPheader->saddr.byte4 == clientIP.byte4)
			{
				DataSize[Count]=dataLen;	
				Data[Count]=malloc(dataLen);
                		memcpy(Data[Count], (u_char *)(pkt_data+layer2Len+IPlen+TCPlen), dataLen);
				seq[Count]=0;
        			Count++;
			}
			if(IPheader->saddr.byte4 == serverIP.byte4)
			{	
				DataSize[Count]=dataLen;	
				Data[Count]=malloc(dataLen);
                		memcpy(Data[Count], (u_char *)(pkt_data+layer2Len+IPlen+TCPlen), dataLen);
				seq[Count]=1;
        			Count++;
			}
		}
        }

	/*Handle UDP*/
        if(IPheader->protocol==0x11)
	{
		///Set UDPheader and calculate data length
                UDPheader = (udp_header *) ((u_char*)IPheader + IPlen);
                UDPlen= 8;
                dataLen=( (header->len)-layer2Len-IPlen-UDPlen);
		
		///For the first time through we set the client & server IP
		///The very FIRST IP will be considered the client
		if ( (firstTime == 0) )
		{
			firstTime = 1;
			clientIP = IPheader->saddr;
			serverIP = IPheader->daddr;
			printf("Found in Capture file:\nclientIP=%d.%d.%d.%d dport=%d --> ",
				clientIP.byte1,clientIP.byte2,clientIP.byte3,clientIP.byte4,ntohs(UDPheader->dport));
			printf("serverIP=%d.%d.%d.%d\n",serverIP.byte1,serverIP.byte2,serverIP.byte3,serverIP.byte4);
			fflush(stdout);
		}
		
		///If there's data put it in the appropriate array
		if ( (dataLen>0) )
		{
			if(IPheader->saddr.byte4 == clientIP.byte4)
			{
				DataSize[Count]=dataLen;	
				Data[Count]=malloc(dataLen);
                		memcpy(Data[Count], (u_char *)(pkt_data+layer2Len+IPlen+UDPlen), dataLen);
				seq[Count]=0;
        			Count++;
			}
			if(IPheader->saddr.byte4 == serverIP.byte4)
			{	
				DataSize[Count]=dataLen;	
				Data[Count]=malloc(dataLen);
                		memcpy(Data[Count], (u_char *)(pkt_data+layer2Len+IPlen+UDPlen), dataLen);
				seq[Count]=1;
        			Count++;
			}
		}
        }

	/*Handle ICMP*/
        if(IPheader->protocol==0x01)
	{
                printf(" ICMP ");
        }
}

/*opens file and calls readin*/
int call_readin(u_char * dumpin, u_char * dumpout)
{
    pcap_t *fp;
    u_char errbuf[PCAP_ERRBUF_SIZE];

    /* Open capture file */
    if ( (fp = pcap_open_offline(dumpin, errbuf) ) == NULL)
    {
        TIME(); fprintf(stderr,"{{{errbuf=%s}}}\n", errbuf);
        return -1;
    }
    /* Open the dump file */
    dumpfile = pcap_dump_open(fp, dumpout);
    if(dumpfile==NULL)
    {
        TIME(); fprintf(stderr,"{{{Error opening output file}}}\n");
        return -1;
    }
    
    /* Loop over the readin function*/
    pcap_loop(fp, 0, readin, NULL);
    
return 0;
}

void TIME()
{
	struct tm *tm;
	u_char buf[100];
	time_t now;

	now = time(NULL);
	tm = localtime(&now);
	if(strftime(buf, sizeof(buf), "%b %d %H:%M:%S", tm) > 0)
		fprintf(stderr,"\nTime of error: %s, Error:", buf);
}	

int BufferAttack(u_int currentLeg, u_int currentPosition, u_int attackSize, u_int attackChar, int connfd, struct addrinfo *res, u_int userFunc)
{
	u_int totalLen, i;
	int ret=0;
	u_char * Packet;
	u_char * half1;
	u_char * half2;

	half1=malloc(totalLen);
	half2=malloc(totalLen);
	memset(half1, 0x00, totalLen);
	memset(half2, 0x00, totalLen);
	///process and construct the attack line
	if(userFunc)
	USERFUNC(0, currentLeg, 2, currentPosition, 0, 0, connfd, res);
	totalLen=attackSize+DataSize[currentLeg];
	Packet=malloc(totalLen);
	memset(Packet, 0x00, totalLen);
	if(currentPosition==0)
	{
		if(attackChar==0x43)
		{
			for(i=0;i<attackSize;i++) { Packet[i]=RAND(256); }
			memcpy( Packet+attackSize, Data[currentLeg], DataSize[currentLeg]);
		}
		else
		{
			memset(Packet, attackChar, attackSize);
			memcpy(Packet+attackSize, Data[currentLeg], DataSize[currentLeg]); 	
		}
	}
	else
	{
		memcpy(half1, Data[currentLeg], currentPosition);
		memcpy(half2, Data[currentLeg]+currentPosition, DataSize[currentLeg]-currentPosition);
		memcpy(Packet, half1, currentPosition);
		if(attackChar==0x43)
		{
			for(i=0;i<attackSize;i++) { Packet[i+currentPosition]=RAND(256); }
			memcpy(Packet+currentPosition+attackSize, half2, DataSize[currentLeg]-currentPosition);
		}
		else
		{
			memset(Packet+currentPosition, attackChar, attackSize);
			memcpy(Packet+currentPosition+attackSize, half2, DataSize[currentLeg]-currentPosition); 	
		}
	}
	///process and print attack
	if(userFunc)
		USERFUNC(0, currentLeg, 3, currentPosition, Packet, totalLen, connfd, res);
	if (currentPosition==0)
	{
		printf("[%d][-%d->]  [b-0]",currentLeg, attackSize);
		PRINT(DataSize[currentLeg], printChoice, Data[currentLeg]);
			printf("\n");
	}
	else
	{
		printf("[%d][-%d->] ",currentLeg, attackSize);
		PRINT(currentPosition, printChoice, half1);
		printf("[b-%d]", currentPosition);
		PRINT(DataSize[currentLeg]-currentPosition, printChoice, half2);
			printf("\n");
	}
	//send attack
	if(WRITE(connfd, Packet, totalLen) < 1) {
		ret=1; TIME(); perror("{{{Buffer attack write error -- quit and loop}}}"); }
	free(Packet);
	free(half1);
	free(half2);
	return ret;
}

int FormatAttack(u_int currentLeg, u_int currentPosition, u_int attackSize, u_int attackChar, int connfd, struct addrinfo *res, u_int userFunc)
{
	u_int totalLen, i;
	int ret=0;
	u_char * Packet;
	u_char * half1;
	u_char * half2;

	half1=malloc(totalLen);
	half2=malloc(totalLen);
	memset(half1, 0x00, totalLen);
	memset(half2, 0x00, totalLen);
	///process and construct the attack line
	if(userFunc)
		USERFUNC(0, currentLeg, 2, currentPosition, 0, 0, connfd, res);
	totalLen=(2*attackSize)+DataSize[currentLeg]; ///Times 2 for the %'s
	Packet=malloc(totalLen);
	memset(Packet, 0x00, totalLen);
	if(currentPosition==0)
	{
		if(attackChar==0x43)
		{
			for(i=0;i<(2*attackSize);i=i+2)
			{
				Packet[i]='%';
				Packet[i+1]=RAND(128); 
			}
			memcpy( Packet+(2*attackSize), Data[currentLeg], DataSize[currentLeg]);
		}
		else
		{
			for(i=0;i<(2*attackSize);i=i+2)
			{
				Packet[i]='%';
				Packet[i+1]=attackChar; 
			}
			memcpy(Packet+(2*attackSize), Data[currentLeg], DataSize[currentLeg]); 	
		}
	}
	else
	{
		memcpy(half1, Data[currentLeg], currentPosition);
		memcpy(half2, Data[currentLeg]+currentPosition, DataSize[currentLeg]-currentPosition);
		memcpy(Packet, half1, currentPosition);
		if(attackChar==0x43)
		{
			for(i=0;i<(2*attackSize);i=i+2)
			{
				Packet[i+currentPosition]='%';
				Packet[i+1+currentPosition]=RAND(128);
			}
			memcpy(Packet+currentPosition+(2*attackSize), half2, DataSize[currentLeg]-currentPosition);
		}
		else
		{
			for(i=0;i<(2*attackSize);i=i+2)
			{
				Packet[i+currentPosition]='%';
				Packet[i+1+currentPosition]=attackChar; 
			}
			memcpy(Packet+currentPosition+(2*attackSize), half2, DataSize[currentLeg]-currentPosition); 	
		}
	}	
	///process, send, and print attack
	if(userFunc)
		USERFUNC(0, currentLeg, 3, currentPosition, Packet, totalLen, connfd, res);
	if (currentPosition==0)
	{
		printf("[%d][-%d->]  [f-0]",currentLeg, attackSize);
		PRINT(DataSize[currentLeg], printChoice, Data[currentLeg]);
			printf("\n");
	}
	else
	{
		printf("[%d][-%d->] ",currentLeg, attackSize);
		PRINT(currentPosition, printChoice, half1);
		printf("[f-%d]", currentPosition);
		PRINT(DataSize[currentLeg]-currentPosition, printChoice, half2);
			printf("\n");
	}
	//send attack
	if(WRITE(connfd, Packet, totalLen) < 1) {
		ret=1; TIME(); perror("{{{Format attack write error -- quit and loop}}}"); }
	free(Packet);
	free(half1);
	free(half2);
	return ret;
}

int LogicAttack(u_int currentLeg, u_int currentPosition, u_int attackSize, u_int attackChar, int connfd, struct addrinfo *res, u_int userFunc)
{
	u_int totalLen, i, k, len;
	int ret=0;
	u_char * Packet;
	///deal with null legs (should that ever happen?)
	if( (len=DataSize[currentLeg]) == 0 )
		return -1;
	///process and construct the attack line
	if(userFunc)
		USERFUNC(0, currentLeg, 2, currentPosition, 0, 0, connfd, res);
	Packet=malloc(len);
	if(attackChar==0x43)
	{
		for(i=0;i<len;i++) { Packet[i]=RAND(256); }
	}
	else
	{
		memcpy(Packet, Data[currentLeg], len);
		if(attackSize<255)
			Packet[currentPosition]=attackSize;
		else
		{
			for(i=RAND(len); i>0; i--)
			{
				k=RAND(len);
				Packet[k]=RAND(256);
			}
		}
	}	
	/// Process, send, and print attack
	totalLen=len;  //done for the sake of isakmp userFunc
	if(userFunc)
		USERFUNC(0, currentLeg, 3, currentPosition, Packet, totalLen, connfd, res);
	printf("[%d][-l_%x->] ",currentLeg, attackSize);
	PRINT(totalLen, printChoice, Packet);
		printf("\n");
	//send attack
	if(WRITE(connfd, Packet, totalLen) < 1) {
		ret=1; TIME(); perror("{{{Logic attack write error -- quit and loop}}}"); }
	free(Packet);
	return ret;
}

//1=Yes. 0=No.
u_int Is_String(struct Tok * tok)
{
	int i;
	u_char c;
	u_int ret=1;
	for(i=0; i<tok->dataLen; i++)
	{
		c=(u_char)*(tok->data+i);
		if ( (c > (u_char)0x80) )
		{
			ret=0;
			break;
		}	
	}
	return ret;
}
			
//1=Yes. 0=No.
u_int Is_Numeric(struct Tok * tok)
{
	int i;
	for(i=0; i<tok->dataLen; i++)
	{
		if ( (tok->data[i] >= 0x30) && (tok->data[i] <=0x39) )
		{
			return 1;
		}	
	}
	return 0;
}
			
//1=Yes. 0=No.
u_int Is_Bracket(struct Tok * tok)
{
	int i;
	u_char c;
	u_int ret=0;
	for(i=0; i<tok->dataLen; i++)
	{
		c=(u_char)*(tok->data+i);
		if ( (c=='{')||(c=='}')||(c=='[')||(c==']')||(c=='(')||(c==')')||(c=='<')||(c=='>') )
		{
			ret=1;
			break;
		}	
	}
	return ret;
}
//1=Yes. 0=No.
u_int Is_Quote(struct Tok * tok)
{
	int i;
	u_char c;
	u_int ret=0;
	for(i=0; i<tok->dataLen; i++)
	{
		c=(u_char)*(tok->data+i);
		if ( (c =='\'')||(c=='\"')||(c=='`') )
		{
			ret=1;
			break;
		}	
	}
	return ret;
}
			
//1=Yes. 0=No.
u_int Is_Special(struct Tok * tok)
{
	int i;
	u_char c;
	u_int ret=0;
	for(i=0; i<tok->dataLen; i++)
	{
		c=(u_char)*(tok->data+i);
		if ( (c=='~')||(c=='!')||(c=='@')||(c=='#')||(c=='$')||(c=='%')||(c=='^')||(c=='&')||(c=='*')||(c=='/')||(c=='\\') )
		{
			ret=1;
			break;
		}	
	}
	return ret;
}
			
//1=Yes. 0=No.
u_int Is_Other(struct Tok * tok)
{
	int i;
	u_char c;
	u_int ret=0;
	for(i=0; i<tok->dataLen; i++)
	{
		c=(u_char)*(tok->data+i);
		if ( (c==',')||(c=='.')||(c=='?')||(c=='|')||(c=='=')||(c=='+')||(c=='-')||(c=='_') )
		{
			ret=1;
			break;
		}	
	}
	return ret;
}
			
//-1=error. 0=ok.
int Build_Fuzzy_Leg(u_char ** Packet, u_int * totalLen, struct Tok * tok, u_int respect, u_int leg)
{
	struct Tok * fuzzyTok;
	struct Tok * startTok;
	struct Tok * currentTok;
	u_int i,x;

	startTok=tok;
	currentTok=tok;
	//lame error checking
	if( (tok->data == NULL) || (tok->dataLen==0) )
		return -1;
	if( respect==0 || respect ==1)
	{
	}
	else
	{
		return -1;
	}
	if( leg < 0)
		return -1;
	//fuzz either according to respect for current leg or not
	if(respect)
	{
		//Find start of current leg
		while(1)
		{
			if(currentTok->leg == leg)
				break;
			else
				currentTok=currentTok->next;
		}
	}
	else
	{
		//ignore leg and randomly pick a tok to start at
		x=RAND(Count);
		for(i=0; i<x; i++)
		{
			currentTok=currentTok->next;
			if(currentTok==NULL)
			{
				currentTok=startTok;
				break;
			}
		}
	}
	while(1)
	{
		//dup good tok to the our fuzzy tok
		fuzzyTok=Dup_Tok(currentTok);
		//fuzz token
		if( Fuzz(fuzzyTok) == -1 )
		{
			TIME(); fprintf(stderr, "{{Fuzz failed!}}\n"); goto END;
		}
		if( (fuzzyTok->string == 1) && (fuzzyTok->legEnd == 1) )
		{
			if( Fuzz_Str_Ending(fuzzyTok) == -1 )
			{
				TIME(); fprintf(stderr, "{{Fuzz_Str_Ending failed!}}\n"); goto END;
			}
		}
		else if ( (fuzzyTok->string == 1) && (RAND(100) < 3) )
		{
			if( Fuzz_Str_Ending(fuzzyTok) == -1 )
			{
				TIME(); fprintf(stderr, "{{Fuzz_Str_Ending failed!}}\n"); goto END;
			}
		}
		//cat fuzzy tok to leg (Packet) and update totalLen (of Packet)
		*Packet=realloc(*Packet, (*totalLen)+(fuzzyTok->dataLen)+(fuzzyTok->delimLen) );
		memcpy( (*Packet)+(*totalLen), fuzzyTok->data, fuzzyTok->dataLen);
		(*totalLen)=(*totalLen)+(fuzzyTok->dataLen);
		//sometimes we add the delimiter back in
		if( (RAND(100) < 96) && (fuzzyTok->legEnd==0) )
		{
			memcpy( (*Packet)+(*totalLen), fuzzyTok->delim, fuzzyTok->delimLen);
			(*totalLen)=(*totalLen)+(fuzzyTok->delimLen);
		}
		END:
		//clean up fuzzyTok memory	
		Destroy_Tok(fuzzyTok);
		//if we're to the end bailout or not depending on respect
		if(respect)
		{
			if(fuzzyTok->legEnd == 1)
				break;
			else
				currentTok=currentTok->next;	
		}
		else
		{
			if( (RAND(100)) < 25 )
				break;
			else
			{
				currentTok=currentTok->next;
				if(currentTok==NULL)
					currentTok=startTok;
			}
		}
	}
}

//-1=error. 0=ok.
int Fuzz(struct Tok * tok)
{
	int ret=0;
	u_int x;

	if( (tok->data == NULL) || (tok->dataLen==0) )
		return -1;
//THIS IS A VERY IMPORTANT FUNC TO HAVE GOOD HEURSITICS/WEIGHTS IN
//x% chance to fuzz token
//Is there a way I can weight state/path commands (i.e. USER) less to fuzz than the data after it?  Would we want to?
//be sure to include error checking around each function
	//than each attack within has it's own % of being invoked
	x=0;
	if( RAND(100) < 30) //% of the time we'll fuzz this token
	{
		//If the buf is a string, fuzz according to our string heuristics
		if( tok->string == 1)
		{
			x=RAND(100);
			if( tok->numeric==1 && (x < 25) )
				ret=Fuzz_Numeric(tok);
			else if (tok->numeric==0 && (x<3) )
				ret=Fuzz_Numeric(tok);
			x=RAND(100);
			if( (tok->bracket==1 || tok->quote==1 || tok->special==1 || tok->other==1) && (x < 25) )
				ret=Fuzz_Parsers(tok);
			else if( (x < 3) )
				ret=Fuzz_Parsers(tok);
			if( RAND(100) < 25 )
				ret=Fuzz_BitFlip(tok);
			if( RAND(100) < 25 )
				ret=Fuzz_Buffer(tok);
			if( RAND(100) < 12 )
				ret=Fuzz_Format(tok);
		}
		//If the buf contines binary data, fuzz according to our binary heuristics
		//I'd like to have a special notion of the size:data format to fuzz size along with data
		//i.e. sometimes we correct and sometimes we don't
		else
		{
		}
	}
	return ret;
}

//-1=error. 0=ok.
//code could be converted to deal with long long int
int Fuzz_Numeric(struct Tok * tok)
{
	u_int i, numLen, fuzzyNumLen, lenDiff, sizeToEnd;
	u_int x=0;
	u_char * numStr=NULL;
	u_char * numEnd=NULL;
	u_char * half2=NULL;
	u_char fuzzyNum[20];
	int num;
	
	if( (tok->data == NULL) || (tok->dataLen==0) )
	{
		fprintf(stderr, "{{error1 in Fuzz_Numeric}}");
		return -1;
	}
	//cp tok->data into new string we can muck with
	if( (numStr=malloc(tok->dataLen+1)) == NULL)
	{
		fprintf(stderr, "{{error2 in Fuzz_Numeric}}");
		return -1;
	}
	memcpy(numStr+1, tok->data, tok->dataLen);
	numStr[0]=0x20; //strtoul needs white space at start of string
	//find start of num
	for(i=0; i<(tok->dataLen+1); i++)
	{
		if ( isdigit(numStr[i]) != 0 )
			break;
	}
	//code to handle the case where no number is found
	if( i == tok->dataLen+1)
	{
		x=RAND(tok->dataLen)+1;
		i=x;
	}
	else //code that deals with a number that was already there
	{
		//back up i if the previous is a +, -, or x (for hex)
		if ( (numStr[i-1] == '+') || (numStr[i-1] == '-') )
			i--;
		//add one white space if there isn't one before i
		if ( isspace(numStr[i-1]) == 0 )
			numStr[i-1]=0x20; //for some lame reason strtoul needs this
	}
	if( x == 0)
	{
		//the special base 0, expects 0x for hex, nothing for 10 or second char of 0 for octal
		num=strtol(numStr+i, (char **)&numEnd, 0);
		if( num==LONG_MIN || num==LONG_MAX )
		{
			perror("{{error3 in Fuzz_Numeric}}");
		}
		//calulate size of current num
		if ( numEnd > (numStr+1+(tok->dataLen)) )
			numEnd=(numStr+1+(tok->dataLen));
		numLen=(numEnd)-(numStr+i);
	}
	else
	{
		num=RAND(10000)-RAND(10000);
		numLen=0;
		numEnd=(numStr+1)+i;
		i++;
	}
	//make a copy of the data after num
	sizeToEnd=(numEnd)-(numStr+1);
	sizeToEnd=tok->dataLen-sizeToEnd;
	num=Fuzz_Integer(num);
	if( RAND(100) < 70 ) //print as decimal
		sprintf(fuzzyNum, "%d", num);
	else
		sprintf(fuzzyNum, "0x%x", num);
	//get new len and fix tok
	fuzzyNumLen=strlen(fuzzyNum);
	lenDiff=fuzzyNumLen-numLen;
	tok->dataLen+=lenDiff;
	tok->data=realloc(tok->data, tok->dataLen);
	memcpy( (tok->data)+(i-1), fuzzyNum, fuzzyNumLen);
	free(numStr);
	if( sizeToEnd > 0)
	{
		half2=malloc(sizeToEnd);
		memcpy(half2, numEnd, sizeToEnd);
		memcpy( (tok->data)+(i-1+fuzzyNumLen), half2, sizeToEnd);
		free(half2);
	}
	return 0;
}

//no error return possibility. This code had better be right.
//It is the users responsibility to check (and print if desired) rather
//or not the returned INT is possitive or negitive
int Fuzz_Integer(int in)
{
	int out, r;
	u_int x, y, z;
	x=RAND(100);
	y=RAND(100);
	z=RAND(100);
	//+ or - somthing
	if( x < 40)
	{
		if(y < 50) //+ or - small change
		{
			if(z < 50)
				out=in-(RAND(10)+1);
			else
				out=in+(RAND(10)+1);
		}
		else if(y < 85) //+ or - med change
		{
			if(z < 50)
				out=in-(RAND(1000)+1);
			else
				out=in+(RAND(1000)+1);
		}
		else //+ or - huge change
		{
			if(z < 50)
				out=in-(RAND(1000000000)+1);
			else
				out=in+(RAND(1000000000)+1);
		}
	}
	//boundry numbers
	else if( x < 80)
	{
		//pick a random exponent (of size int) for 2^r
		r=RAND(33);
		out=pow(2, r);
		if (y < 33) //+ or - 1
		{
			if(z < 50)
				out++;
			else
				out--;
		}
		else if (y < 66) //+ or - 2
		{
			if(z < 50)
				out+=2;
			else
				out-=2;
		}
		else //leave at boundry
		{
		}
	}
	//pick a wild number
	else
	{
		if (y < 33)
			out=RAND(1000)-RAND(1000); //give a small + or -
		else if (y < 66)
			out=RAND(100000)-RAND(100000); //give a med + or -
		else
			out=RAND(RAND_MAX)+RAND(RAND_MAX); //give a huge + or -
	}
	
	return out;
}

//-1=error. 0=ok.
int Fuzz_Parsers(struct Tok * tok)
{
	int i, x;
	u_char array[]="{}[]()<>\'\"`~!@#$%^&*/\\,.?|=+-_{}[]()<>\'\"";//added a few twice, why??
	u_int parsersNum=0;
	void * parsersPtr[300]; //ugly
	u_char * newPkt=NULL;
	u_char * nextPtr=NULL;
	u_char * prevPtr=NULL;
	u_char * fuzzyPtr=NULL;
	u_char * pkt;
	u_char byte;
	u_int newPktLen=0;
	u_int fuzzyLen=0;

	if( (tok->data == NULL) || (tok->dataLen==0) )
	{
		fprintf(stderr, "{{error1 in Fuzz_Parsers}}");
		return -1;
	}
	//Find the position and number of all parsers
	for(i=0; i<tok->dataLen; i++)
	{
		if( (strchr(array, tok->data[i])) !=NULL)
		{
			parsersPtr[parsersNum]=(tok->data)+i;
			parsersNum++;
		}	
	}
	//code to handel if no parsers
	if(parsersNum == 0)
	{
		pkt=malloc(tok->dataLen+2);
		memcpy(pkt+1, tok->data, tok->dataLen);
		byte=array[RAND(strlen(array))];
		*(pkt)=byte;
		*(pkt+tok->dataLen+1)=byte;
		free(tok->data);
		tok->data=pkt;
		tok->dataLen+=2;

		parsersPtr[0]=tok->data;
		parsersPtr[1]=(tok->data)+tok->dataLen-1;
		parsersNum=2;
	}
	//Fuzz each parser
	prevPtr=tok->data;
	for(i=0; i<parsersNum; i++)
	{
		nextPtr=parsersPtr[i];
		x=nextPtr-prevPtr;
		//make room data
		if(x>0)
		{
			newPkt=realloc(newPkt, newPktLen+x);
			//first copy up to parser
			memcpy(newPkt+newPktLen, prevPtr, x);
			newPktLen+=x;
		}
		//fuzz
		if( (fuzzyPtr=_Fuzz_Parsers(array, *(nextPtr))) == NULL)
		{
			fprintf(stderr, "{{error4 in Fuzz_Parsers}}");
			return -1;
		}
		//get fuzzy len
		fuzzyLen=strlen(fuzzyPtr);
		//make room for fuzzy
		if(fuzzyLen>0)
		{
			newPkt=realloc(newPkt, newPktLen+fuzzyLen);
			//cp in new data and start again
			memcpy(newPkt+newPktLen, fuzzyPtr, fuzzyLen);
			newPktLen+=fuzzyLen;
		}
		free(fuzzyPtr);
		prevPtr=nextPtr;
		prevPtr++;
	}
	//tack the tail on
	nextPtr=(tok->data)+(tok->dataLen);
	x=nextPtr-prevPtr;
	//make room data
	if(x >0)
	{
		newPkt=realloc(newPkt, newPktLen+x);
		memcpy(newPkt+newPktLen, prevPtr, x);
		newPktLen=newPktLen+x;
	}
	//clean up and reset ptrs
	free(tok->data);
	tok->data=newPkt;
	tok->dataLen=newPktLen;
	return 0;
}

//array must be a vaild string
//original is use as a bias toward
//what the parser was (i.e. sometimes it'll be unchanged).
//NULL=error. User strlen for len.
//the string MUST be freed by the user.
u_char * _Fuzz_Parsers(u_char * array, u_char original)
{
	char * out;
	u_int x, y, z, i, arrayLen;
	x=RAND(100);
	y=RAND(100);
	if( array==NULL )
	{
		fprintf(stderr, "{{error1 in _Fuzz_Parsers}}");
		return NULL;
	}
	if( (arrayLen=strlen(array)) == 0)
	{
		fprintf(stderr, "{{error2 in _Fuzz_Parsers}}");
		return NULL;
	}

	if( x < 35) //empty string
	{
		out=malloc(1+1);
		if( y < 40)
		{
			out[0]=0x00;
		}
		else
		{
			out[0]=original;
		}
		out[1]=0x00;
	}
	else if( x < 70) //one
	{
		out=malloc(1+1);
		out[0]=array[RAND(arrayLen)];
		out[1]=0x00;
	}
	else if( x < 83) //1-10, diff or same
	{
		z=RAND(10)+1;
		out=malloc(z+1);
		if( y < 65)
			memset(out, array[RAND(arrayLen)], z);
		else
		{
			for(i=0; i<z; i++)
				out[i]=array[RAND(arrayLen)];
		}
		out[z]=0x00;
	}
	else if( x < 96) //1-100, diff or same
	{
		z=RAND(100)+1;
		out=malloc(z+1);
		if( y < 65)
			memset(out, array[RAND(arrayLen)], z);
		else
		{
			for(i=0; i<z; i++)
				out[i]=array[RAND(arrayLen)];
		}
		out[z]=0x00;
	}
	else //1-10000, diff or same
	{
		z=RAND(10000)+1;
		out=malloc(z+1);
		if( y < 65)
			memset(out, array[RAND(arrayLen)], z);
		else
		{
			for(i=0; i<z; i++)
				out[i]=array[RAND(arrayLen)];
		}
		out[z]=0x00;
	}
	return out;
}

//-1=error. 0=ok.
int Fuzz_BitFlip(struct Tok * tok)
{
	u_int i, r;
	u_int num;

	if( (tok->data == NULL) || (tok->dataLen==0) )
		return -1;
	//randomly pick a mode
	//we either want to flip just a couple bits of one byte
	if( (r=RAND(100)) < 40)
	{
		num=tok->dataLen;
		tok->data[RAND(num)]=tok->data[RAND(num)] & (RAND(256));
	}
	//or randomly replace a couple bytes
	else if( r < 80)
	{
		num=RAND(2)+1;
		for(i=0;i<num;i++)
			tok->data[RAND(tok->dataLen)]=RAND(256);	
	}
	//or randomly replacy (rand(dataLen)) bytes
	else
	{
		num=RAND(tok->dataLen);
		for(i=0;i<num;i++)
			tok->data[RAND(tok->dataLen)]=RAND(256);	
	}
	return 0;
}

//-1=error. 0=ok.
int Fuzz_Buffer(struct Tok * tok)
{
	u_int i, r;
	char * half2;
	u_int position, attackSize;
	u_char c;

	if( (tok->data == NULL) || (tok->dataLen==0) )
		return -1;
	//randomly pick a position to insert format attack
	position=RAND(tok->dataLen+1);
	//randomly pick an attackSize 
	if( (r=RAND(100)) < 40)
		attackSize=RAND(30);
	else if( r < 93)
		attackSize=RAND(3000);
	else
		attackSize=RAND(100000);
	//configure to half
	half2=malloc((tok->dataLen)-position);
	memcpy(half2, (tok->data)+position, (tok->dataLen)-position);
	if( ((tok->data)=realloc((tok->data), ((attackSize)+(tok->dataLen)))) == NULL)
	{
		TIME(); fprintf(stderr, "{{realloc in Fuzz_Format failed!}}\n"); return (-1);
	}
	//create and insert attack
	if( (r=RAND(100)) < 15) //all chars
	{
		for(i=0;i<attackSize;i++)
			*((tok->data)+(i+position))=RAND(256);
		memcpy((tok->data)+position+(attackSize), half2, (tok->dataLen)-position);
	}
	else if (r < 35 )  //ascii only
	{
		for(i=0;i<attackSize;i++)
			*((tok->data)+(i+position))=RAND(128);
		memcpy((tok->data)+position+(attackSize), half2, (tok->dataLen)-position);
	}
	else if (r < 70) //only alphanumeric
	{
		for(i=0;i<attackSize;i++)
			*((tok->data)+(i+position))=RAND(96)+32;
		memcpy((tok->data)+position+(attackSize), half2, (tok->dataLen)-position);
	}
	else  //only one char 
	{
		c=RAND(96)+32;
		for(i=0;i<attackSize;i++)
			*((tok->data)+(i+position))=c;
		memcpy((tok->data)+position+(attackSize), half2, (tok->dataLen)-position);
	}
	tok->dataLen=(tok->dataLen)+(attackSize);
	free(half2);
	return 0;
}

//-1=error. 0=ok.
int Fuzz_Format(struct Tok * tok)
{
	u_int i, x;
	u_char c;
	char * half2;
	u_int position, attackSize;

	if( (tok->data == NULL) || (tok->dataLen==0) )
		return -1;
	//randomly pick a position to insert format attack
	position=RAND(tok->dataLen+1);
	//randomly pick an attackSize 
	attackSize=RAND(30);
	//configure to half
	half2=malloc((tok->dataLen)-position);
	memcpy(half2, (tok->data)+position, (tok->dataLen)-position);
	if( ((tok->data)=realloc((tok->data), ((2*attackSize)+(tok->dataLen)))) == NULL)
	{
		TIME(); fprintf(stderr, "{{realloc in Fuzz_Format failed!}}\n"); return (-1);
	}
	//create and insert attack
	if( RAND(100) < 15)
	{
		for(i=0;i<(2*attackSize);i=i+2)
		{
			*((tok->data)+(i+position))='%';
			*((tok->data)+(i+1+position))=RAND(128);
		}
		memcpy((tok->data)+position+(2*attackSize), half2, (tok->dataLen)-position);
	}
	else
	{
		if( (x=RAND(100)) < 20)
			c='s';
		else if( x < 40)
			c='d';
		else if( x < 60)
			c='x';
		else
			c='n';
		for(i=0;i<(2*attackSize);i=i+2)
		{
			*((tok->data)+(i+position))='%';
			*((tok->data)+(i+1+position))=c; 
		}
		memcpy((tok->data)+position+(2*attackSize), half2, (tok->dataLen)-position); 	
	}
	tok->dataLen=(tok->dataLen)+(2*attackSize);
	free(half2);
	return 0;
}

//-1=error. 0=ok.
int Fuzz_Str_Ending(struct Tok * tok)
{
	u_int r=0;

	if( (tok->data == NULL) || (tok->dataLen==0) )
		return -1;
	if( (tok->data=realloc(tok->data, tok->dataLen+2)) == NULL )
	{
		TIME(); fprintf(stderr, "{{realloc in Fuzz_Str_Ending failed!}}\n"); return (-1);
	}
	//trunk line
	if( tok->data[tok->dataLen-2] == '\r')
		tok->dataLen=tok->dataLen-2;
	else if( tok->data[tok->dataLen-3] == '\r')
		tok->dataLen=tok->dataLen-3;
	else if( tok->data[tok->dataLen-1] == '\n')
		tok->dataLen=tok->dataLen-1;
	else if( tok->data[tok->dataLen-2] == '\n')
		tok->dataLen=tok->dataLen-2;
	else //tack on an ending where there was none
	{
	}
	//randomize which ending we'll tack-on
	r=RAND(101);
	if( r < 70)
	{
		memcpy(tok->data+tok->dataLen, "\r\n", 2);
		tok->dataLen=tok->dataLen+2;
	}
	else if( r < 76)
	{
		memcpy(tok->data+tok->dataLen, "\n\r", 2);
		tok->dataLen=tok->dataLen+2;
	}
	else if( r < 82)
	{
		memcpy(tok->data+tok->dataLen, "\n", 1);
		tok->dataLen++;
	}
	else if( r < 88)
	{
		memcpy(tok->data+tok->dataLen, "\r", 1);
		tok->dataLen++;
	}
	else if( r < 94)
	{
		memcpy(tok->data+tok->dataLen, "\x00", 1);
		tok->dataLen++;
	}
	//none
	else
	{
	}
	return 0;
}

void * Create_Tok()
{
	struct Tok * tok;
	tok=malloc(sizeof(struct Tok));

	tok->dataLen=0;
	tok->data=NULL;
	tok->string=0;
	tok->numeric=0;
	tok->bracket=0;
	tok->quote=0;
	tok->special=0;
	tok->other=0;
	tok->delimLen=0;
	tok->delim=NULL;
	tok->leg=0;
	tok->legEnd=0;
	tok->next=NULL;

	return tok;
}

void Destroy_Tok(struct Tok * tok)
{
	free(tok->delim);
	free(tok->data);
	free(tok);
}

void * Dup_Tok(struct Tok * tokin)
{
	struct Tok * tokout;
	tokout=malloc(sizeof(struct Tok));

	tokout->dataLen=tokin->dataLen;
	tokout->data=malloc(tokout->dataLen);
	memcpy(tokout->data, tokin->data, tokout->dataLen);
	tokout->string=tokin->string;
	tokout->numeric=tokin->numeric;
	tokout->bracket=tokin->bracket;
	tokout->quote=tokin->quote;
	tokout->special=tokin->special;
	tokout->other=tokin->other;
	tokout->delimLen=tokin->delimLen;
	tokout->delim=malloc(tokout->delimLen);
	memcpy(tokout->delim, tokin->delim, tokout->delimLen);
	tokout->leg=tokin->leg;
	tokout->legEnd=tokin->legEnd;
	tokout->next=tokin->next;

	return tokout;
}
//num="1" to print just the tok you pass
//num="2" for that one and the next, ...
//num="all" to print all toks
void Print_Toks(struct Tok * tok, char * num)
{
	u_int i, _num;

	if( (strcmp("all", num)) == 0)
		_num=5000;
	else
		_num=atoi(num);
	for(i=0; i<_num; i++)
	{
		printf("tok:%x\n", tok);
		printf("\ttok->dataLen=%d\n", tok->dataLen);
		printf("\ttok->data=");
		PRINT(tok->dataLen, 2, tok->data);
			printf("\n");
		printf("\ttok->string=%d\n", tok->string);
		printf("\ttok->numeric=%d\n", tok->numeric);
		printf("\ttok->bracket=%d\n", tok->bracket);
		printf("\ttok->quote=%d\n", tok->quote);
		printf("\ttok->special=%d\n", tok->special);
		printf("\ttok->other=%d\n", tok->other);
		printf("\ttok->delimLen=%d\n", tok->delimLen);
		printf("\ttok->delimeter=");
		if(tok->delim == NULL)
			printf("NULL\n");
		else
		{
			PRINT(tok->delimLen, 2, tok->delim);
				printf("\n");
		}
		printf("\ttok->leg=%d\n", tok->leg);
		printf("\ttok->legEnd=%d\n", tok->legEnd);
		printf("\ttok->next=%x\n", tok->next);
		if( tok->next ==  NULL)
			break;
		else
			tok=tok->next;
	}
}
